/*------------------------------------------------------------------------------
   EPA - Region II, ESD
  ------------------------------------------------------------------------------
   Program: eyton .c
   Purpose: calculate Eyton's ellipse classification with INFO data file input
  ------------------------------------------------------------------------------
   Usage:   see char *usage
  ------------------------------------------------------------------------------
   Calls:   statistics library
   Calling: eyton.aml
  ------------------------------------------------------------------------------
   Notes:   
  ------------------------------------------------------------------------------
   History:
     Author/Site:    Date:       Event:
   Alan Brenner    16 Nov 92   Original coding
                   12 Jan 94   update
		   17 Aug 95   changed standard deviation function
     alan_brenner@unixmail.edi.epa.gov
  ------------------------------------------------------------------------------
   Disclaimer:
     Although this program has been used by the U. S. Environmental Protection
   Agency, no warranty, expressed or implied, is made by the EPA as to the
   accurarcy and functioning of the program and related program material nor
   shall the fact of distribution constitute any such warranty, and no
   responsibility is assumed by the EPA in connection herewith.
==============================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "infolib.h"
#include "infodefs.h"
#include "utils.h"
#include "stats.h"

char debug=0;						    /* debugger */

/* determine if the point is inside or outside the  chi-squared ellipse =========================================================*/
double ellipse(x, y, sx, sy, mx, my, r)
  double x,y,sx,sy,mx,my,r;
{							    /* from J. Ronald Eyton, 1984 */
     double sxy,xmx,ymy;				    /* _Complementary-Color, Two Variable Maps_ */
							    /* The Annals of the Association of American Geographers */
     sxy=sx*sy;						    /* 74(3) 477-490 */
     if (sxy==0)
	  error(4,"Floating point error in ellipse.");
     sx*=sx;
     if (sx==0)
	  error(5,"Floating point error in ellipse.");
     sy*=sy;
     if (sy==0)
	  error(6,"Floating point error in ellipse.");

     xmx=x-mx; ymy=y-my;

     return((xmx*xmx/sx+ymy*ymy/sy)/(1-r*r)-2*r*xmx*ymy/sxy);
}

/* main function ================================================================================================================*/
int main (argc, argv)
  int argc;
  char *argv[];
{
     short ewidth,eoutput,edecimals,*red,*cyan;
     long etype,rows,classes,count,counted,*counter,z;
     double chisq,mx,my,sx,sy,r,t,u,minx,maxx,miny,maxy,*x,*y;
     char input[85];
     InfoFile *infofile;
     InfoItemDef *xitem, *yitem, *eitem, *reditem, *cyanitem, *countitem;


     if (argc<5)
	  error(-1,"Usage: eyton <data_table> <lookup_table> <x_item> <y_item> {2|3|4|6} {chi-square}");

/* get data --------------------------------------------------------------------------------------------------------------------*/
     if ((infofile=InfoOpenFile(argv[1],InfoWRITE))==(InfoFile*)NULL)
	  error(7,"Could not open info file.");
     rows=InfoNumberRecords(infofile);

     if ((xitem=InfoGetItemDef(infofile,argv[3]))==(InfoItemDef*)NULL)
	  error(7,"Could not open x-item.");

     if ((yitem=InfoGetItemDef(infofile,argv[4]))==(InfoItemDef*)NULL)
	  error(7,"Could not open y-item.");

     if ((eitem=InfoGetItemDef(infofile,"EYTON"))==(InfoItemDef*)NULL)
	  error(7,"Could not open output item EYTON.");
     ewidth=InfoItemWidth(eitem);
     eoutput=InfoItemOutputWidth(eitem);
     etype=InfoItemType(eitem);
     edecimals=InfoItemNumberDecimals(eitem);

     if (argc>5)
     {
	  if (*(argv[5])!=50 && *(argv[5])!=51 && *(argv[5])!=52 && *(argv[5])!=54)
	       classes=2;				    /* default is 2x2 box for non-polygons */
	  else
	       classes=atol(argv[5]);			    /*gets number of classes from command line*/
     }
     else
	  classes=2;					    /* default if no item */

     if (argc>6)
	  chisq=atof(argv[6]);				    /*get chi-squared value*/
     else
	  chisq=1.386;					    /*50% of data in the ellipse, by default*/
     if (!(chisq>0))
	  error(8,"Chi-squared must be greather than 0.");

     x=calloc(rows,sizeof(double));			    /*get data space*/
     y=calloc(rows,sizeof(double));

/* get data --------------------------------------------------------------------------------------------------------------------*/
     if (classes==2 || classes==3)			    /* non-polygon attribute table */
	  counted=0;					    /* read all of the data fields */
     else
     {							    /* polygon attribute tables only */
	  counted=1;					    /* skip the universe polygon */
	  classes/=2;					    /* reset the number of classes */
	  rows--;					    /* decrease the number of rows by one */
     }
     for (count=0; count<rows; count++)			    /* read data from info */
     {
	  if (InfoReadRecord(infofile,count+1+counted)==INFO_FAIL)
	       error(9,"Error reading info file.");	    /* offset by counted to skip universe polygon */
	  if (InfoDecode(infofile,xitem,input,x+count)==INFO_FAIL)
	       error(9,"Error decoding x item.");
	  if (InfoDecode(infofile,yitem,input,y+count)==INFO_FAIL)
	       error(9,"Error decoding y item.");

     }

     red=calloc(classes*classes+1,sizeof(short));	    /* get space for lookup table data */
     cyan=calloc(classes*classes+1,sizeof(short));
     counter=calloc(classes*classes+1,sizeof(long));
     for (count=0; count<classes*classes+1; count++) *(counter+count)=0;

/* find classification parameters ----------------------------------------------------------------------------------------------*/
     minx=maxx=*x;
     miny=maxy=*y;
     for (count=counted+1; count<rows; count++)		    /*find data extremes*/
     {
	  if (x[count]<minx) minx=x[count];
	  if (x[count]>maxx) maxx=x[count];
	  if (y[count]<miny) miny=y[count];
	  if (y[count]>maxy) maxy=y[count];
     }

     mx=mean(rows,x);					    /* find arithmetic means */
     my=mean(rows,y);
     sx=ssd(rows,mx,x);					    /* find standard deviations */
     sy=ssd(rows,my,y);
     r=pearson(rows,mx,my,x,y,&t,&u);			    /* find correlation */
     if (r==0)
	  error(3,"Zero correlation.");
     if (r*r==1)
	  error(3,"Perfect linear correlation.");
     (void) printf("\nCoefficient of correlation (Pearson's r) = %f.\n",r);
     (void) printf("Student's t probability for r = %f (lower is better).\n\n",t);

/* find classification for each point-------------------------------------------------------------------------------------------*/
     for (count=0; count<rows; count++)			    /*for each data pair*/
     {
	  if (chisq<ellipse(x[count],y[count],sx,sy,mx,my,r)) /*outside of ellipse?*/
	  {
	       if (classes==2)				    /*2x2 square*/
	       {
		    if (x[count]<(minx+maxx)/2.0)
		    {
			 if (y[count]<(miny+maxy)/2.0) z=2;
			 else z=3;
		    }
		    else
		    {
			 if (y[count]<(miny+maxy)/2.0) z=5;
			 else z=4;
		    }
	       }
	       else					    /*3x3 square*/
	       {
		    if(x[count]<(minx+maxx)/3.0)
		    {
			 if (y[count]<(miny+maxy)/3.0) z=2;
			 else if (y[count]>(miny+maxy)*2.0/3.0) z=4;
			 else z=3;
		    }
		    else if (x[count]>(minx+maxx)*2.0/3.0)
		    {
			 if (y[count]<(miny+maxy)/3.0) z=8;
			 else if (y[count]>(miny+maxy)*2.0/3.0) z=6;
			 else z=7;
		    }
		    else
		    {
			 if (y[count]<(miny+maxy)/3.0) z=9;
			 else if (y[count]>(miny+maxy)*2.0/3.0) z=5;
			 else z=10;
		    }
	       }
	  }
	  else z=1;					    /*middle gray*/
	  counter[z-1]++;

	  if (InfoReadRecord(infofile,count+1+counted)==INFO_FAIL)
	       error(10,"Error reading info file.");	    /* read record */
	  if (InfoEncode(infofile,eitem,input,(double)z)==INFO_FAIL)
	       error(10,"Error encoding classification.");  /* change eyton score */
	  if (InfoWriteRecord(infofile,count+1+counted)==INFO_FAIL)
	       error(10,"Error writing info file.");	    /* write record--not just the item!!! */
     }

     if (InfoFileFlush(infofile)==INFO_FAIL)		    /* flush and close info file */
	  error(10,"Error flushing info file.");
     if (InfoCloseFile(infofile)==INFO_FAIL)
	  error(10,"Error closing info file.");
     
/* set color scheme variables --------------------------------------------------------------------------------------------------*/
     red[0]=40; cyan[0]=40;				    /* set colors--ellipse as middle gray */
     if (r<0)						    /* negative correlation */
     {
	  red[1]=100; cyan[1]=0;			    /* lower left is red */
	  if (classes==2)
	  {
	       red[2]=10;  cyan[2]=10;			    /* upper left is light gray */
	       red[3]=0;   cyan[3]=100;			    /* upper right is cyan */
	       red[4]=100; cyan[4]=100;			    /* lower right is black */
	  }
	  else
	  {
	       red[2]=40;  cyan[2]=10;			    /* center left */
	       red[3]=10;  cyan[3]=10;			    /* upper left is light gray */
	       red[4]=10;  cyan[4]=40;			    /* upper center */
	       red[5]=0;   cyan[5]=100;			    /* upper right is cyan*/
	       red[6]=40;  cyan[6]=100;			    /* center right */
	       red[7]=100; cyan[7]=100;			    /* lower right is black */
	       red[8]=100; cyan[8]=40;			    /* lower center */
	       red[9]=60;  cyan[9]=60;			    /* behing elipse is dark gray */
	  }
     }
     else						    /* positive correlation */
     {
	  red[1]=10; cyan[1]=10;			    /* lower left is light gray */
	  if (classes==2)
	  {
	       red[2]=100; cyan[2]=0;			    /* upper left is red */
	       red[3]=100; cyan[3]=100;			    /* upper right is black */
	       red[4]=0;   cyan[4]=100;			    /* lower right is cyan */
	  }
	  else
	  {
	       red[2]=40;  cyan[2]=10;			    /* center left */
	       red[3]=100; cyan[3]=0;			    /* upper left is red */
	       red[4]=100; cyan[4]=40;			    /* upper center */
	       red[5]=100; cyan[5]=100;			    /* upper right is black */
	       red[6]=40;  cyan[6]=100;			    /* center right */
	       red[7]=0;   cyan[7]=100;			    /* lower right is cyan */
	       red[8]=10;  cyan[8]=40;			    /* lower center */
	       red[9]=60;  cyan[9]=60;			    /* behind elipse is dark gray */
	  }
     }

/* create output lookup table --------------------------------------------------------------------------------------------------*/
     if (InfoCreateFile(argv[2],(InfoFile*)NULL)==INFO_FAIL)
	  error(11,"Could not create lookup table.");	    /* create lookup table */
     
     if (InfoDefineItem(argv[2],"EYTON",ewidth,eoutput,etype,edecimals)==INFO_FAIL)
	  error(11,"Could not define EYTON in lookup table.");
     
     if (InfoDefineItem(argv[2],"RED",4,5,INFO_BINARY_TYPE,0)==INFO_FAIL)
	  error(11,"Could not define RED in lookup table.");

     if (InfoDefineItem(argv[2],"CYAN",4,5,INFO_BINARY_TYPE,0)==INFO_FAIL)
	  error(11,"Could not define CYAN in lookup table.");

     if (InfoDefineItem(argv[2],"COUNT",4,5,INFO_BINARY_TYPE,0)==INFO_FAIL)
	  error(11,"Could not define COUNT in lookup table.");

     if (InfoDefineItem(argv[2],"NOM",40,40,INFO_CHARACTER_TYPE,0)==INFO_FAIL)
	  error(11,"Could not define NOM in lookup table.");

     if ((infofile=InfoOpenFile(argv[2],InfoWRITE))==(InfoFile*)NULL)
	  error(12,"Could not open info file.");	    /* open lookup table */

     if ((eitem=InfoGetItemDef(infofile,"EYTON"))==(InfoItemDef*)NULL)
	  error(12,"Could not open EYTON lookup item.");
     
     if ((reditem=InfoGetItemDef(infofile,"RED"))==(InfoItemDef*)NULL)
	  error(12,"Could not open RED lookup item.");
     
     if ((cyanitem=InfoGetItemDef(infofile,"CYAN"))==(InfoItemDef*)NULL)
	  error(12,"Could not open CYAN lookup item.");
     
     if ((countitem=InfoGetItemDef(infofile,"COUNT"))==(InfoItemDef*)NULL)
	  error(12,"Could not open COUNT lookup item.");
     
     for (count=0; count<classes*classes+1; count++)	    /*write data for lookup table*/
     {
	  if (InfoEncode(infofile,eitem,input,(double)(count+1))==INFO_FAIL)
	       error(13,"Error encoding eyton item.");
	  if (InfoEncode(infofile,reditem,input,(double)red[count])==INFO_FAIL)
	       error(13,"Error encoding red item.");
	  if (InfoEncode(infofile,cyanitem,input,(double)cyan[count])==INFO_FAIL)
	       error(13,"Error encoding cyan item.");
	  if (InfoEncode(infofile,countitem,input,(double)*(counter+count))==INFO_FAIL)
	       error(13,"Error encoding count.");

	  if (InfoWriteRecord(infofile,count+1)==INFO_FAIL)
	       error(13,"Error writing lookup table.");
     }

     if (InfoFileFlush(infofile)==INFO_FAIL)		    /* flush and close info file */
	  error(14,"Error flushing info file.");
     if (InfoCloseFile(infofile)==INFO_FAIL)
	  error(14,"Error closing info file.");

     return(0);						    /*return no errors*/
}

