/*------------------------------------------------------------------------------
   EPA - Region II, ESD
  ------------------------------------------------------------------------------
   Program: jenks .c
   Purpose: calculate Jenk's Optimal classification with INFO data file input
  ------------------------------------------------------------------------------
   Usage:   see char *usage
  ------------------------------------------------------------------------------
   Calls:   jenkslib
   Calling: jenks.aml
  ------------------------------------------------------------------------------
   Notes:   
  ------------------------------------------------------------------------------
   History:
     Author/Site:    Date:       Event:
   Alan Brenner    21 Jul 93   Original coding
                   11 Jan 94   update
		   16 Aug 95   reworked jenkshort routines
		               joined functions that dir (not) handle weights
     alan_brenner@unixmail.edi.epa.gov
  ------------------------------------------------------------------------------
   Disclaimer:
     Although this program has been used by the U. S. Environmental Protection
   Agency, no warranty, expressed or implied, is made by the EPA as to the
   accurarcy and functioning of the program and related program material nor
   shall the fact of distribution constitute any such warranty, and no
   responsibility is assumed by the EPA in connection herewith.
==============================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "infolib.h"
#include "infodefs.h"
#include "utils.h"
#include "inforeselect.h"
#include "jenkslib.h"					    /* this is the wieghted (duplicate values) jenks library */

char    debug=0;                                            /* debugger */

/* routine for quicksort of testpnts structures =================================================================================*/
int structsort(one, two)
  struct jtest *one;
  struct jtest *two;
{
     int choice=0;

     if (one->pnt<two->pnt) choice=-1;
     if (one->pnt>two->pnt) choice=1;

     return(choice);
}

/* do a quick estimation of the best breaks =====================================================================================*/
double jshort(class, cgvf, sdam, bestp, breaks, points, hits)
  const long class;					    /* number of classes */
  double cgvf;						    /* inital/current goodness of variance fit */
  const double sdam;					    /* initial sum of the deviations of the array mean */
  long *bestp;						    /* best break points */
  long *breaks;						    /* current break points */
  double *points;					    /* data values */
  const unsigned long *hits;				    /* number of instances of each value in points or NULL */
{
     long size;						    /* class squared shorthand */

     struct jtest *testpnts;				    /* array of breakpoint/size info */
     long mingap=0;					    /* record number of minimum gap in testpnts */
     long freegaps=0;					    /* number of gaps of same size removed from testpnts */

     double diff;					    /* current difference */

     long count;					    /* for loop counter */

     int (*choose)()=structsort;			    /* qsort order function */

     size=class*class;					    /* set shorthand */

     testpnts=calloc(size,sizeof(jteststruct));		    /* space for big gaps */
     for (count=0; count<size; count++)
     {
	  testpnts[count].pnt=count;			    /* initialize to first size breaks */
	  if (hits==NULL)				    /* unweighted gap */
	       testpnts[count].gap=points[count+1]-points[count];
	  else						    /* weighted gap */
	       testpnts[count].gap=(hits[count+1]*points[count+1])-(hits[count]*points[count]);
	  if (testpnts[count].gap<testpnts[mingap].gap)	    /* if this gap is smaller than the last smallest */
	       mingap=count;				    /* remember this as the smallest */
     }

/* find the class squared biggest gaps -----------------------------------------------------------------------------------------*/
     for (count=size; count<breaks[class]-1; count++)	    /* for every data value not already set */
     {
	  if (hits==NULL)				    /* find the size of the gap between data values */
	       diff=points[count+1]-points[count];	    /* unweighted */
	  else						    /* weighted */
	       diff=(hits[count+1]*points[count+1])-(hits[count]*points[count]);

	  if (testpnts[mingap].gap==diff)
	  {						    /* duplicate size gap */
	       long counter;				    /* for loop counter */

	       if (freegaps==0)				    /* no blanks in testpnts */
	       {					    /* so remove the minimum */
		    for (counter=0; counter<size; counter++)
		    {
			 if (testpnts[counter].gap==diff)
			 {
			      testpnts[counter].pnt=-2;
			      testpnts[counter].gap=0.0;
			      freegaps++;
			 }
		    }
	       }
	       else					    /* insert this into a blank */
	       {
		    for (counter=0; counter<size; counter++)
		    {
			 if (testpnts[counter].gap==-2 && testpnts[counter].gap==0.0)
			 {
			      testpnts[counter].pnt=count;  /* remember data */
			      testpnts[counter].gap=diff;
			      freegaps--;		    /* one less free space */
			      break;			    /* exit for loop */
			 }
		    }
	       }
	  }

	  if (testpnts[mingap].gap<diff)
	  {						    /* minimum gap is smaller than this one */
	       testpnts[mingap].pnt=count;		    /* remember this record number */
	       testpnts[mingap].gap=diff;
	  }
     }

/* sort the data and find the most significant of the class squared gaps -------------------------------------------------------*/
     qsort(testpnts,size,sizeof(jteststruct),choose);	    /* sort data */

     testpnts=testpnts+freegaps;			    /* move beginning of array to first read data value */
     for (count=1; count<class; count++)
     {
	  breaks[count]=testpnts[count-1].pnt;		    /* load first points into breaks */
     }

     return(step(class, class-1, cgvf, sdam, bestp, breaks, points, hits, NULL, size-freegaps, testpnts, freegaps));
}

/* main routine ================================================================================================================*/
int main(argc, argv)
  int argc;
  char *argv[];
{
     char *usage="Usage: jenks(hort) <data_table> <lookup_table> <item> <classes> {polygons}";
     char jname=0;					    /* name of program: 0=jenks 1=jenkshort */

     int (*choose)()=dpick;				    /* qsort order function */

     long classes;					    /* number of classes to put data into */

     long count;					    /* for loop counter */
     long counter;					    /* for loop counter nested in count */

     long rows;						    /* number of data items */
     long row;						    /* number of data items after duplicates removed */
     long rowed;					    /* estimate of significance of row */

     long *breakp;					    /* array of the current test breakpoints */
     long *bestb;					    /* array of best breakpoint found */

     double sdam;					    /* sum of the deviations of the whole array */
     double gvf=0.0;					    /* value of the goodness of variance fit for the best breaks */

     double *points1;					    /* input data values */
     double *points2;					    /* data values after duplicates noted in hits and removed */
     unsigned long *hits;				    /* array of the number of occurance of each value in points2 */

     InfoFile *infofile;				    /* input data file */
     char *ctmp="";					    /* dummy character item for getting numbers from INFO table */
     long offset=0;					    /* record number offset */
     InfoItemDef *dataitem;				    /* data item in infofile */
     short iwidth;					    /* internal width of dataitem */
     short ioutput;					    /* display width of dataitem */
     short idecimals;					    /* number of decimal place in dataitem */
     long itype;					    /* storage type of dataitem */
     unsigned long InfoReselect=0;                          /* number of reselected items--default is 0 */
     InfoReselectStruct *InfoReselects=NULL;                /* pointer to empty reselect list */

     InfoItemDef *symitem;				    /* output table items */
     InfoItemDef *countitem;
     unsigned long total;				    /* number of data values in each class */
     double next;					    /* output temp values */
     double last;

/* command line switches -------------------------------------------------------------------------------------------------------*/
     if (strncmp(argv[0]+strlen(argv[0])-10,"inforegexd",10)==0)
          debug=9;                                          /* set debug output to max if called as inforegexd */

     if (debug > 8)                                         /* print command line parameters */
     {
          BUGOUT(stderr,"\nSTART\n");
          for (count=0; count<argc; count++)
               BUGOUT(stderr,"arg %d=%s\n",count,argv[count]);
     }

     if (strncmp("jenkshort",argv[0],9)==0)
	  jname=1;					    /* called as jenkshort */
     NEXTARG;
     
     while ((argc>0) && (argv[0][0]=='-'))                  /* if there are arguements and the first character is a - */
     {
          switch (argv[0][1])
          {
             case 'd':                                      /* set the debug value */
               if (strlen(argv[0])>2)                       /* if there is something after the d */
               {
                    if (isdigit(argv[0][2])!=0)             /* if it is a digit */
                         debug=(char) atoi(&argv[0][2]);    /* set the debug value to it */
                    else
                         debug=1;                           /* otherwise minimal debuging */
		    NEXTARG;
               }
               else                                         /* chack the next agruement */
               {
                    if (argc>1)                             /* if there is another argument, */
                    {
			 NEXTARG;
                         if (isdigit(argv[0][0])!=0)        /* is it a number? */
                         {
                              debug=(char) atoi(argv[2]);   /* if yes, set the debug level to it */
			      NEXTARG;
                         }
			 else
			      debug=1;			    /* not a digit, set minimal debug */
                    }
               }
               break;

             case 'r':
               if (strlen(argv[0])>2)                       /* if there is something after the r */
	       {
                    InfoReselect=InfoGetReselects(&(argv[0][2]),&InfoReselects);
		    NEXTARG;
	       }
               else                                         /* check the next agruement */
               {
                    if (argc>1)                             /* is there another argument? */
                    {
			 NEXTARG;
                         if (isdigit(argv[0][0])!=0)        /* is it a number? */
                         {
			      InfoReselect=InfoGetReselects(argv[1],&InfoReselects);
			      NEXTARG;
			 }
                    }
               }
               break;

             default:                                       /* check usage */
               error(1,usage);
          }
     }

     if (debug > 8)                                         /* print remaining command line arguments */
     {
          for (count=0; count<argc; count++)
               BUGOUT(stderr,"arg %d=%s\n",count,argv[count]);
          BUGOUT(stderr,"check for file\n");
     }
     if (argc<5) error(1,usage);                            /* not enough arguments after switches */

/* get the data ----------------------------------------------------------------------------------------------------------------*/
     classes=atoi(argv[4]);				    /*gets number of classes from command line*/

     if (classes<2)					    /*checks for screwy input*/
	  error(1,"The number of classes must be greater than one.");

     if ((infofile=InfoOpenFile(argv[1],InfoREAD))==(InfoFile*)NULL)
	  error(2,"Could not open input info file.");	    /* open data file */

     rows=InfoNumberRecords(infofile);			    /* number of data records */

     if ((dataitem=InfoGetItemDef(infofile,argv[3]))==(InfoItemDef*)NULL)
	  error(3,"Could not open data item.");		    /* find data item */

     iwidth=InfoItemWidth(dataitem);
     ioutput=InfoItemOutputWidth(dataitem);
     itype=InfoItemType(dataitem);
     idecimals=InfoItemNumberDecimals(dataitem);

     if (argc>5) offset=1;				    /* if polygon, skip the universe */

     points1=calloc(rows-offset,sizeof(double));	    /* get data space */
     points2=calloc(rows-offset,sizeof(double));
     hits=calloc(rows-offset,sizeof(short));

     for (count=offset+1; count<=rows; count++)		    /* read data */
     {
	  *(hits+count)=1;				    /* all counts start at one */

	  if (InfoReadRecord(infofile,count)==INFO_FAIL)
	       error(4,"Error reading info file.");	    /* read record */

	  if (InfoDecode(infofile,dataitem,ctmp,points1+count)==INFO_FAIL)
	       error(4,"Error decoding info item.");	    /* extract data item */
     }

     if (InfoCloseFile(infofile)==INFO_FAIL)
          error(4,"Error closing info file.");

     makelut(argv[2],argv[3],iwidth,ioutput,itype,idecimals); /* create lookup table--done here to avoid doing the work */
							    /* and then find an INFO limitation for the output (999 tables) */

/* sort the data and determine if there are duplicate values -------------------------------------------------------------------*/
     qsort(points1,rows-offset,sizeof(double),choose);	    /* sort data, then check for duplicate values*/

     row=rows-offset;					    /* assume points1 will have no duplicate values */
     points2[0]=points1[0];				    /* first value is always new to short list */
     for (counter=0,count=1; count<rows-offset; count++)    /* for each data value */
     {
	  if (points2[counter]==points1[count])		    /* current and next are the same */
	  {
	       hits[counter]++;				    /* number at value increments one */
	       row--;					    /* one less points2 row */
	  }
	  else						    /* new value */
	  {
	       counter++;				    /* point to next unique data value */
	       points2[counter]=points1[count];		    /* copy this data value */
	  }
     }

     if (classes-1>=row)				    /*checks for screwy input*/
	  error(5,"The number of classes must be smaller than the number of different data points.");

/* set up the initial break points ---------------------------------------------------------------------------------------------*/
     breakp=calloc(((classes+1)*classes),sizeof(long));	    /*memory for pointers to breakpoints*/
     bestb=calloc(classes+1,sizeof(long));		    /*memory for best breakpoints*/

     if (jname && classes*classes<row)			    /* if the squared number of classes is greater than the number of */
	  jname=0;					    /* data values, than run the ehaustive routine rather than the short*/
     if (jname)						    /* if the program was run as jenkshort */
	  rowed=rows-offset;				    /* note ANY duplicate values */
     else
	  rowed=(rows-offset)*0.9;			    /* otherwise only note >10% duplicates */

     breakp[0]=bestb[0]=-1;				    /* first break is before first item */
     if (row<rowed)
	  breakp[classes]=bestb[classes]=row-1;		    /* end of weighted list */
     else
     {
	  breakp[classes]=bestb[classes]=rows-offset-1;	    /* end of raw list */
	  free(points2);
	  free(hits);					    /* release memery */
	  hits=NULL;					    /* point to NULL */
     }

     for (count=1; count<classes; count++)		    /* set initial breakpoints */
     {							    /* point to data value before break */
	  if (jname)					    /* if the program was run as jenkshort, use equal intervals */
	       bestb[count]=count*bestb[classes]/classes;
	  else
	       bestb[count]=breakp[count]=count-1;	    /* if the exhaustive routine is used, point to low end values */
     }

/* find the best break points --------------------------------------------------------------------------------------------------*/
     if (row<rowed)
	  sdam=ssd(rows-offset,points2,hits,NULL);	    /* deviation for whole array, with weights */
     else
	  sdam=ssd(rows-offset,points1,hits,NULL);	    /* deviation for whole array, without weights */
     if (sdam==0)
	  error(2,"Zero variability--Cannot calculate breakpoints.");

     if (jname)						    /* if the program was run as jenkshort, run quickly */
     {							    /* if data has any duplicate values */
	  if (row<rows-offset)				    /* find the best gvf with integer wieghting */
	       gvf=jshort(classes,gvf,sdam,bestb,breakp,points2,hits);
	  else						    /* find the best gvf without weights */
	       gvf=jshort(classes,gvf,sdam,bestb,breakp,points1,hits);
     }
     else						    /* if the program was run as jenks run exhaustively */
     {							    /* this will take longer, but find the best solution */
	  if (row<rowed)				    /* if data has many (>10%) duplicate values use weights */
	       gvf=step(classes,classes-1,gvf,sdam,bestb,breakp,points2,hits,NULL,0,NULL,0);
	  else
	       gvf=step(classes,classes-1,gvf,sdam,bestb,breakp,points1,NULL,NULL,0,NULL,0);
     }

     (void) printf("\n\n\nThe Goodness of Variance Fit for %d classes = %f.\n\n",classes,gvf);

/* output the results ----------------------------------------------------------------------------------------------------------*/
     openlut(&infofile,argv[2],&dataitem,argv[3],&symitem,&countitem); /* open lookup table */

     last=points2[0]-1;					    /*first line is one smaller than smallest*/
     writelut(infofile,dataitem,last,symitem,(double)0,countitem,(double)(rows-offset),1);

     for (count=1; count<classes; count++)		    /* for each breakpoint */
     {
	  if (row<rowed)				    /* breaks are mean of value pointed at & following */
	       next=(points2[bestb[count]]+points2[bestb[count]+1])/2;
	  else
	       next=(points1[bestb[count]]+points1[bestb[count]+1])/2;

	  for (total=counter=0; counter<row; counter++)	    /*find number of values in class*/
	  {
	       if (row<rowed)
	       {
		    if (points2[counter]>last && points2[counter]<next) total+=hits[counter];
	       }
	       else
	       {
		    if (points1[counter]>last && points1[counter]<next) total++;
	       }
	  }
	  writelut(infofile,dataitem,next,symitem,(double)count,countitem,(double)total,count+1);
	  last=next;
     }

     for (total=counter=0; counter<row; counter++)	    /*find number of values in last class*/
     {
	  if (row<rowed)
	  {
	       if (points2[counter]>last && points1[counter]<points2[row-1]+1) total+=hits[counter];
	  }
	  else
	  {
	       if (points1[counter]>last && points1[counter]<points1[row-1]+1) total++;
	  }
     }
     writelut(infofile,dataitem,points1[rows-offset-1]+1,symitem,(double)(rows-offset),countitem,(double)total,classes+1);

     if (InfoFileFlush(infofile)==INFO_FAIL)                /* flush and close info file */
          error(14,"Error flushing info file.");
     if (InfoCloseFile(infofile)==INFO_FAIL)
          error(14,"Error closing info file.");

     return(0);						    /*return no errors*/
}
