/*------------------------------------------------------------------------------
   EPA - Region II, ESD
  ------------------------------------------------------------------------------
   Program: jenkslib.c
   Purpose: routines used by both jenks.c and weightedjenks.c
  ------------------------------------------------------------------------------
   Usage:   
  ------------------------------------------------------------------------------
   Calls:   
   Calling: jenks.c, weightedjenks.c
  ------------------------------------------------------------------------------
   Notes:   
  ------------------------------------------------------------------------------
   History:
     Author/Site:    Date:       Event:
   Alan Brenner    21 Jul 93   Original coding
                   11 Jan 94   update
                   16 Aug 93   joined functions like sdc1 and sdc2 into sdc
     alan_brenner@unixmail.edi.epa.gov
  ------------------------------------------------------------------------------
   Disclaimer:
     Although this program has been used by the U. S. Environmental Protection
   Agency, no warranty, expressed or implied, is made by the EPA as to the
   accurarcy and functioning of the program and related program material nor
   shall the fact of distribution constitute any such warranty, and no
   responsibility is assumed by the EPA in connection herewith.
==============================================================================*/

#include "jenkslib.h"
#include "utils.h"

extern char    debug;					    /* debugger */

/* calculate sum of squared deviations ==========================================================================================*/
double ssd(rows, data, vals, wats)
  const long rows;					    /* number of data items */
  const double *data;					    /* pointer to array of data values */
  const unsigned long *vals;				    /* number of occurances of each value in points or NULL */
  const double *wats;					    /* weight of each value in points or NULL*/
{
     long count;					    /* for loop counter */
     double items=0.0;					    /* number of data items */
     double ave;					    /* average */
     double sum=0.0;					    /* summation */
     double ssd=0.0;					    /* sum of squared deviations */

     for (count=0; count<rows; count++)			    /* first find the summation of all data values */
     {
	  if (vals==(const unsigned long *)NULL && wats==(const double *)NULL)
	       sum+=data[count];			    /* without weights */
	  else
	  {
	       if (wats==(const double *)NULL)
		    sum+=(data[count]*vals[count]);	    /* with counts */
	       else
		    sum+=(data[count]*wats[count]);	    /* with weights */
	  }
     }

     if (vals==(const unsigned long *)NULL && wats==(const double *)NULL) /* then find the number of instances */
	  items=rows;					    /* unweighted */
     else
     {
	  for (count=0; count<rows; count++)
	  {
	       if (wats==(const double *)NULL)
		    items+=vals[count];			    /* number of actual data items */
	       else
		    items+=wats[count];			    /* sum of weights */
	  }
     }

     ave=sum/(double) items;				    /* calculate mean */

     for (count=0; count<rows; count++)			    /* calculate sum of squared deviations */
     {
	  if (vals==(const unsigned long *)NULL && wats==(const double *)NULL)
	       ssd+=pow(data[count]-ave,2);		    /* unweighted */
	  else
	  {
	       if (wats==(const double *)NULL)		    /* counted */
		    ssd+=vals[count]*pow(data[count]-ave,2);
	       else					    /* weighted */
		    ssd+=wats[count]*pow(data[count]-ave,2);
	  }
     }
     return(ssd);					    /* return ssd */
}

/* calculate sum of deviations of each class ====================================================================================*/
double sdc(classes, breakp, points, vals, wats)
  const long classes;					    /* number of classes*/
  long *breakp;						    /* pointer to array of break points */
  const double *points;					    /* pointer to data values */
  const unsigned long *vals;				    /* number of occurance of each value in points or NULL */
  const double *wats;					    /* weight of each value in points or NULL*/
{
     short count;					    /* for loop counter */
     double sdcm=0.0;					    /* sum of deviations of the class mean */

     for (count=0; count<classes; count++)		    /* loop sums devation of each class */
     {
	  if (vals==(const unsigned long *)NULL && wats==(const double *)NULL)
	       sdcm+=ssd(breakp[count+1]-breakp[count],points+breakp[count]+1,vals,wats);
	  else
	  {
	       if (wats==(const double *)NULL)
		    sdcm+=ssd(breakp[count+1]-breakp[count],points+breakp[count]+1,vals+breakp[count]+1,wats);
	       else
		    sdcm+=ssd(breakp[count+1]-breakp[count],points+breakp[count]+1,vals,wats+breakp[count]+1);
	  }
     }
     return(sdcm);
}

/* determine best goodness of variance fit ======================================================================================*/
double bestgvf(classes, gvfnow, sdam, bestb, breaks, points, vals, wats)
  long *bestb;						    /* array of the best break points */
  const long classes;					    /* number of classes */
  long *breaks;						    /* array of the curret break pointer*/
  double gvfnow;					    /* value of the goodness of variance for best break points */
  const double sdam;					    /* sum of deviation of the array means for current break points */
  double *points;					    /* data values */
  const unsigned long *vals;				    /* number of occurance of each value in points or NULL */
  const double *wats;					    /* weight of each value in points or NULL*/
{
     long count;					    /* for loop counter */
     double gvf;					    /* current goodness of variance fit */

     gvf=(sdam-sdc(classes,breaks,points,vals,wats))/sdam;  /*find gvf with current breaks*/
     if (gvf>gvfnow)					    /*if new gvf is better than former best*/
     {
	  gvfnow=gvf;					    /*copy current to best*/
	  for (count=1; count<classes; count++) bestb[count]=breaks[count];
     }
     return(gvfnow);					    /*return current gvf value*/
}

/* recursively go through every possible break ==================================================================================*/
double step(class, steps, cgvf, sdam, bestp, breaks, points, vals, wats, size, testpnts, testval)
  const long class;					    /* number of classes */
  const long steps;					    /* number of pointers to be moved in this recursion level */
  double cgvf;						    /* value of the goodness of variance for best break points */
  const double sdam;					    /* sum of deviations from the whole array's mean */
  long *bestp;						    /* best breaks */
  long *breaks;						    /* current breaks */
  double *points;					    /* data values */
  const unsigned long *vals;				    /* number of occurance of each value in points or NULL */
  const double *wats;					    /* weight of each value in points or NULL*/
  long size;						    /* number of data values pointed to in testpnts or 0 */
  struct jtest *testpnts;				    /* list of breakpoints to be tested or NULL */
  long testval;						    /* testpnt value to use */
{
     long count;					    /* for loop counters */
     long counter;

     long *broke;					    /* array of current break points */

     long start;					    /* for loop start value */
     long stop;						    /* for loop stop value */

     if (debug>4) BUGOUT(stderr,"step %d %d %f %f %p %p %p %p %p %d %p\n",
			 class, steps, cgvf, sdam, bestp, breaks, points, vals, wats, size, testpnts, testval);

     broke=breaks+(long)((class-steps)*(class+1));	    /* local copy of current break points */
     for (count=0; count<=class; count++)
     {
	  broke[count]=breaks[count];
	  if (debug>6) BUGOUT(stderr,"\t\tbroke[%d]=%d\n",count,broke[count]);
     }

     if (size==0)					    /* set up start and stop values for for loop */
     {
	  start=broke[class-steps];			    /* number of first in list of markers to move at this level */
	  stop=broke[class]-steps;			    /* last data value minus the number of breaks left to move */
     }
     else
     {
	  start=testval;				    /* first point to move at this level */
	  stop=size-steps;				    /* number of pointers to move to before last pointer position */
     }
     if (debug>5) BUGOUT(stderr,"\tstart=%d\tstop=%d\n",start,stop);

     for (count=start; count<=stop; count++)		    /* for each break point to move at this level */
     {							    /* this level handles more than last breakpoint */
	  if (steps>1)					    /* move one less break at next level */
	       cgvf=step(class,steps-1,cgvf,sdam,bestp,broke,points,vals,wats,size,testpnts,testval+1);
	  else						    /* find best vs current */
	       cgvf=bestgvf(class,cgvf,sdam,bestp,broke,points,vals,wats);

	  for (counter=class-steps; counter>=class-1; counter++)
	  {						    /* move this level's breaks */
	       if (size==0)
		    broke[counter]++;			    /* to the next data value */
	       else
	       {
		    broke[counter]=testpnts[testval].pnt;   /* to the next test value */
		    testval++;
	       }
	  }
	  testval+=class;
     }
     return(cgvf);
}

/*==============================================================================================================================*/
void makelut(infoname,itemname,iwidth,ioutput,itype,idecimals)
  char *infoname,*itemname;
  short iwidth,ioutput,idecimals;
  long itype;
{							    /* create lookup table */
     if (InfoCreateFile(infoname,(InfoFile*)NULL)==INFO_FAIL)
          error(1,"Could not create lookup table.");

     if (InfoDefineItem(infoname,itemname,iwidth,ioutput,itype,idecimals)==INFO_FAIL)
          error(1,"Could not define relate item in lookup table.");
     
     if (InfoDefineItem(infoname,"SYMBOL",4,5,INFO_BINARY_TYPE,0)==INFO_FAIL)
          error(1,"Could not define SYMBOL in lookup table.");

     if (InfoDefineItem(infoname,"COUNT",4,5,INFO_BINARY_TYPE,0)==INFO_FAIL)
          error(1,"Could not define COUNT in lookup table.");

     if (InfoDefineItem(infoname,"NOM",40,40,INFO_CHARACTER_TYPE,0)==INFO_FAIL)
          error(1,"Could not define NOM in lookup table.");
}

/*==============================================================================================================================*/
void openlut(infofile,infoname,dataitem,itemname,symitem,countitem)
  char *infoname,*itemname;
  InfoFile **infofile;
  InfoItemDef **dataitem,**countitem,**symitem;
{							    /* open lookup table (AML must write NOM item) */
     if ((*infofile=InfoOpenFile(infoname,InfoWRITE))==(InfoFile*)NULL)
          error(1,"Could not open lookup table.");

     if ((*dataitem=InfoGetItemDef(*infofile,itemname))==(InfoItemDef*)NULL)
          error(1,"Could not open relate lookup item.");
     
     if ((*symitem=InfoGetItemDef(*infofile,"SYMBOL"))==(InfoItemDef*)NULL)
          error(1,"Could not open SYMBOL lookup item.");
     
     if ((*countitem=InfoGetItemDef(*infofile,"COUNT"))==(InfoItemDef*)NULL)
          error(1,"Could not open COUNT lookup item.");
}

/*==============================================================================================================================*/
void writelut(infofile,dataitem,dataval,symitem,symval,countitem,countval,recnum)
  long recnum;
  double dataval,symval,countval;
  InfoFile *infofile;
  InfoItemDef *dataitem,*countitem,*symitem;
{
     if (InfoEncode(infofile,dataitem,NULL,dataval)==INFO_FAIL)
	  error(1,"Error encoding relate item.");
     if (InfoEncode(infofile,symitem,NULL,symval)==INFO_FAIL)
	  error(1,"Error encoding SYMBOL item in lookup table.");
     if (InfoEncode(infofile,countitem,NULL,countval)==INFO_FAIL)
	  error(1,"Error encoding count.");
     if (InfoWriteRecord(infofile,recnum)==INFO_FAIL)
	  error(1,"Error writing lookup table.");
}




