/*---------------------------------------------------------------------------------------------------------------------------------
   EPA - Region II, ESD
  ------------------------------------------------------------------------------
   Program: spear(), crank()
   Purpose: find Spearmans' r
  ------------------------------------------------------------------------------
   Usage:   see function
  ------------------------------------------------------------------------------
   Calls:   ssortdd()
   Calling: general function
  ------------------------------------------------------------------------------
   Notes:   statistical functions some simple and some derived from:
William H. Press, Brian P. Flannery, Saul A. Teukolsky and William T. Vettering
    _NUMERICAL RECIPIES in C: The Art of Scientific Computing_
    Cambridge University Press, 1988
            This book is useful, but contains problems--most notably incorrect
	    subscripting of arrays (C is 0 to N-1; they use 1 to N).
  ------------------------------------------------------------------------------
   History:
     Author/Site:    Date:       Event:
   Alan Brenner      Jan 94      original coding
     alan_brenner@unixmail.edi.epa.gov
   Alan Brenner      Jan 95      added case based statistics
                                 added debuging
  ------------------------------------------------------------------------------
   Disclaimer:
     Although this program has been used by the U. S. Environmental Protection
   Agency, no warranty, expressed or implied, is made by the EPA as to the
   accurarcy and functioning of the program and related program material nor
   shall the fact of distribution constitute any such warranty, and no
   responsibility is assumed by the EPA in connection herewith.
---------------------------------------------------------------------------------------------------------------------------------*/

#include "stats.h"

double spear(n, data1, data2, d, zd, probd, probrs)
  long n;
  double *data1,*data2,*d,*zd,*probd,*probrs;
{
     int j;
     double vard,t,sg,sf,fac,en3n,en,df,aved,*wksp1,*wksp2,rs;
     double crank();

     if (debug) (void) fprintf(stderr,"spear(%d,%p,%p,%p,%p,%p,%p)\n",n,data1,data2,d,zd,probd,probrs);

     if ((wksp1=calloc((unsigned int) n,sizeof(double)))==NULL)
          error(-1,"Unable to allocate memory in function spear.");
     if ((wksp2=calloc((unsigned int) n,sizeof(double)))==NULL)
          error(-2,"Unable to allocate memory in function spear.");
     if (debug>3) (void) fprintf(stderr,"memory allocated\n");

     for (j=0; j<n; j++)
     {
	  wksp1[j]=data1[j];
	  wksp2[j]=data2[j];
	  if (debug>3) fprintf(stderr,"%f\t%f\t%f\t%f\n",wksp1[j],data1[j],wksp2[j],data2[j]);
     }
     if (debug>3) (void) fprintf(stderr,"data copied to local set\n");

     ssortdd(n,wksp1,wksp2);
     sf=crank(n,wksp1);
     ssortdd(n,wksp2,wksp1);
     sg=crank(n,wksp2);
     if (debug>3) (void) fprintf(stderr,"data sorted and ranked\n");

     *d=0.0;
     for (j=0; j<n; j++)
	  *d+=(wksp1[j]-wksp2[2])*(wksp1[j]-wksp2[j]);
     en=(double) n;
     en3n=en*en*en-en;
     aved=en3n/6.0-(sf+sg)/12.0;
     fac=(1.0-sf/en3n)*(1.0-sg/en3n);
     vard=((en-1.0)*en*en*((en+1.0)*(en+1.0))/36.0)*fac;
     *zd=(*d-aved)/sqrt(vard);
     rs=(1.0-(6.0/en3n)*(*d+(sf+sg)/12.0))/sqrt(fac);
     fac=(rs+1.0)*(1.0-rs);
     if (fac>0.0)
     {
	  t=rs*sqrt((en-2.0)/fac);
	  df=en-2.0;
	  *probrs=betai(df/2.0,0.5,df/(df+t*t));
     }
     else
	  *probrs=0.0;
     free(wksp1);
     free(wksp2);
     *probd=erfcc(fabs(*zd)/sqrt(2.0));

     if (debug>3) (void) fprintf(stderr,"\nspear(%d,%p,%p,%d,%d,%d,%d)\n",n,data1,data2,*d,*zd,*probd,*probrs);
     return(rs);
}

double crank(n, data)
  long n;
  double *data;
{
     long j=0;
     double s=0.0;

     if (debug>6)
     {
	  long i;
	  (void) fprintf(stderr,"crank %d\n",n);
	  for (i=0; i<n; i++) fprintf(stderr,"%d\t%f\n",i,data[i]);
     }

     while (j<n-1)					    /* for all except last data element */
     {
	  if (data[j]!=data[j+1])			    /* different data values */
	  {
	       if (debug>6) (void) fprintf(stderr,"%16.6f %11.0d +1=%11.0d\trank %6.0d\n",data[j],j,j+1,j+1);
	       data[j]=(double) j+1;			    /* data value replaced by current count */
	       j++;					    /* increment count */
	  }
	  else						    /* duplicate data values */
	  {
	       long l,k=2;				    /* number of duplicates is at least 2 */
	       double rank;
	       
	       for (; j+k<n && data[j]==data[j+k]; k++);    /* determine number of duplicates */
	       rank=(double) j + (double) (k+1)/2.0;	    /* rank is the mean of the range of equal values */
	       if (debug>6)
		    (void) fprintf(stderr,"%16.6f %5.0d-%5.0d +1=%5.0d-%5.0d\trank %6.1f",data[j],j,j+k-1,j+1,j+k,rank);
	       for (l=0; l<k; l++) data[j+l]=rank;	    /* replace data with rank */
	       s+=k*k*k-k;
	       j+=k;
	       if (debug>6)
		    (void) fprintf(stderr,"\tk=%d\ts=%f\tj=%d\n",k,s,j);
	  }
     }
     if (j==n-1) data[n-1]=n;

     if (debug>5)
     {
	  long i;
	  for (i=0; i<n; i++) fprintf(stderr,"%d\t%f\n",i,data[i]);
     }

     return(s);
}
